/*

 ============================================================================
                   The Apache Software License, Version 1.1
 ============================================================================
 
 Copyright (C) @year@ The Apache Software Foundation. All rights reserved.
 
 Redistribution and use in source and binary forms, with or without modifica-
 tion, are permitted provided that the following conditions are met:
 
 1. Redistributions of  source code must  retain the above copyright  notice,
    this list of conditions and the following disclaimer.
 
 2. Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
 
 3. The end-user documentation included with the redistribution, if any, must
    include  the following  acknowledgment:  "This product includes  software
    developed  by the  Apache Software Foundation  (http://www.apache.org/)."
    Alternately, this  acknowledgment may  appear in the software itself,  if
    and wherever such third-party acknowledgments normally appear.
 
 4. The names "Jakarta", "Apache Avalon", "Avalon Excalibur", "Avalon
    Framework" and "Apache Software Foundation"  must not be used to endorse
    or promote products derived  from this  software without  prior written
    permission. For written permission, please contact apache@apache.org.
 
 5. Products  derived from this software may not  be called "Apache", nor may
    "Apache" appear  in their name,  without prior written permission  of the
    Apache Software Foundation.
 
 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 APACHE SOFTWARE  FOUNDATION  OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL  DAMAGES (INCLU-
 DING, BUT NOT LIMITED TO, PROCUREMENT  OF SUBSTITUTE GOODS OR SERVICES; LOSS
 OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
 This software  consists of voluntary contributions made  by many individuals
 on  behalf of the Apache Software  Foundation and was  originally created by
 Stefano Mazzocchi  <stefano@apache.org>. For more  information on the Apache 
 Software Foundation, please see <http://www.apache.org/>.
 
*/
package org.apache.framework.component.impl;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import org.apache.framework.component.ComponentLocator;
import org.apache.framework.component.ComponentException;

/**
 * This class is a static implementation of a ComponentLocator. Allow ineritance
 * and extension so you can generate a tree of ComponentLocator each defining
 * Object scope.
 *
 * @author <a href="mailto:fede@apache.org">Federico Barbieri</a>
 * @author <a href="mailto:peter@apache.org">Peter Donald</a>
 * @author <a href="mailto:mcconnell@apache.org">Stephen McConnell</a>
 * @version 1.0
 */
public class DefaultComponentLocator
    implements ComponentLocator
{
    private final HashMap               m_objects = new HashMap();
    private final ComponentLocator      m_parent;
    private boolean                     m_readOnly;

    /**
     * Construct ComponentLocator with no parent.
     *
     */
    public DefaultComponentLocator()
    {
        this( null );
    }

    /**
     * Construct ComponentLocator with specified parent.
     *
     * @param parent the ComponentLocator parent
     */
    public DefaultComponentLocator( final ComponentLocator parent )
    {
        m_parent = parent;
    }

    /**
     * Retrieve Object by role from ComponentLocator.
     *
     * @param role the role
     * @return the Object
     * @exception org.apache.framework.component.ComponentException if an error occurs
     */
    public Object lookup( final String role )
        throws ComponentException
    {
        return lookup( role, null );
    }

    /**
     * Retrieve Object by role/hint from ComponentLocator.
     *
     * @param role the role
     * @return the Object
     * @exception org.apache.framework.component.ComponentException if an error occurs
     */
    public Object lookup( final String role, Object hint )
        throws ComponentException
    {
        // the default version of the method ignores the hint
        final Object object = m_objects.get( role );

        if( null != object )
        {
            return object;
        }
        else if( null != m_parent )
        {
            return m_parent.lookup( role );
        }
        else
        {
            throw new ComponentException( role, "Unable to provide implementation." );
        }
    }

    /**
     * Check to see if a <code>Object</code> exists for a role.
     *
     * @param role  a string identifying the role to check.
     * @return True if the object exists, False if it does not.
     */
    public boolean exists( final String role )
    {
        boolean objectExists = false;

        try
        {
            this.lookup(role);
            objectExists = true;
        }
        catch (Throwable t)
        {
            // Ignore all throwables--we want a yes or no answer.
        }
        return objectExists;
    }

    /**
     * Place Object into ComponentLocator.
     *
     * @param role the components role
     * @param object an <code>Object</code> value
     */
    public void put( final String role, final Object object )
    {
        checkWriteable();
        m_objects.put( role, object );
    }

    /**
     * Build a human readable representation of the ComponentLocator.
     *
     * @return the description of the ComponentLocator
     */
    public String toString()
    {
        final StringBuffer buffer = new StringBuffer();
        final Iterator objects = m_objects.keySet().iterator();
        buffer.append( "Services:" );

        while( objects.hasNext() )
        {
            buffer.append( "[" );
            buffer.append( objects.next() );
            buffer.append( "]" );
        }

        return buffer.toString();
    }

    /**
     * Helper method for subclasses to retrieve parent.
     *
     * @return the parent ComponentLocator
     */
    protected final ComponentLocator getParent()
    {
        return m_parent;
    }

    /**
     * Helper method for subclasses to retrieve object map.
     *
     * @return the object map
     */
    protected final Map getObjectMap()
    {
        return m_objects;
    }

    /**
     * Makes this service manager read-only.
     *
     */
    public void makeReadOnly()
    {
        m_readOnly = true;
    }

    /**
     * Checks if this service manager is writeable.
     *
     * @exception java.lang.IllegalStateException if this service manager is read-only
     */
    protected final void checkWriteable()
        throws IllegalStateException
    {
        if( m_readOnly )
        {
            throw new IllegalStateException
                ( "ComponentLocator is read only and can not be modified" );
        }
    }

    
}
